#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/soundcard.h>

#define STRINGLEN		0xff
#define VERSION			"r0.0.1"

#define DEFAULT_MIXER_FILE 	"/dev/mixer"
#define DEFAULT_DSP_FILE	"/dev/dsp"

#define AUDIO_BUFFER_SIZE	4096

#define AC_SETSAMPLERATE 	(1<<0)
#define AC_SETBITRATE		(1<<1)
#define AC_DISPINFO		(1<<2)
#define AC_DISPHELP		(1<<3)
#define AC_DISPVOLS		(1<<4)
#define AC_ADDRECSRC		(1<<5)
#define AC_REMRECSRC		(1<<6)
#define AC_SETRECSRC		(1<<7)
#define AC_SETVOLUMES		(1<<8)
#define AC_RECORD		(1<<9)
#define AC_PLAY			(1<<10)
#define AC_SETSTEREO		(1<<11)
#define AC_DURATION		(1<<12)

struct setreq_t
  {
    int devnumber;
    char left, right;
  };

struct setreq_t vol_requests[SOUND_MIXER_NRDEVICES];

static char *usage_string =
"usage: %s [options]\n";
static char *help_string =
"\n"
"   -h, --help                  : displays this help message\n"
"   -i, --information           : displays information (lot) on the device\n"
"   -v, --volumes               : displays left/right volumes of all devices\n"
"   -c, --record                : continuous read from the dsp, dumped to file\n"
"   -l, --play                  : continuous write to the dsp, read from file\n"
"   -e, --stereo                : set the signal processor in stereo mode\n"
"   -s, --samplerate=VALUE      : sets the signal processor's speed\n"
"   -b, --bitrate=VALUE         : sets the length of signals (8 or 16)\n"
"   -d, --duration=VALUE        : duration of record/play action, by default in\n"
"                                 seconds or in bytes if followed by a 'b'\n"
"   -a, --addsource=STRING      : add a recording source\n"
"   -r, --remsource=STRING      : removes a recording source\n"
"   -m, --mixer=STRING          : changes mixer file (default=/dev/mixer)\n"
"   -p, --dsp=STRING            : changes dsp file (default=/dev/dsp)\n"
"   -t, --setvolume=STRING      : set a volume's device to value (which is\n"
"                                 formatted as 'device-left:right')\n"
"   -f, --file=SRTING           : set the value of file for --record and --play\n"
"                                 Its default value is respectively 'stdin' and\n"
"                                 'stdout' for these two functions.\n"
"\n";


static struct option long_options[] =
{
  {"help", 0, 0, 'h'},
  {"information", 0, 0, 'i'},
  {"volumes", 0, 0, 'v'},
  {"record", 0, 0, 'c'},
  {"play", 0, 0, 'l'},
  {"stereo", 0, 0, 'e'},
  {"samplerate", 1, 0, 's'},
  {"bitrate", 1, 0, 'b'},
  {"duration", 1, 0, 'd'},
  {"addsource", 1, 0, 'a'},
  {"remsource", 1, 0, 'r'},
  {"mixer", 1, 0, 'm'},
  {"dsp", 1, 0, 'p'},
  {"setvolume", 1, 0, 't'},
  {"file", 1, 0, 'f'},
  {NULL, 0, 0, 0}
};

int oper_running = 1;		// set to nul on caught sigterm

int audio_fd, mixer_fd;
int optindex = 0, value = 0;
int samplerate = 0, bitrate = 0;
int reqsamplerate = 0, reqbitrate = 0;
int reqaddsrcmask = 0, reqremsrcmask = 0;
int devmask, srcmask, recmask;
int count = 0, count2 = 0, total = 0;
int tleft, tright;
char *device_names[] = SOUND_DEVICE_NAMES;
char mixer_filename[STRINGLEN], dsp_filename[STRINGLEN];
char tempstring[STRINGLEN], tempstring2[STRINGLEN];
char oper_filename[STRINGLEN];
char *audio_buffer;
int audio_buffer_size = AUDIO_BUFFER_SIZE;
int oper_fd = 0;
long actions = 0;
int stereo = 0;
float durationsecs = 0;
double durationbytes = 0;

void 
signal_handler (int signal)
{
  oper_running = 0;
}

int 
safe_open (char *name, int flags)
{
  int t_fd;
  if ((t_fd = open (name, flags)) < 0)
    {
      printf ("error: couldn't open %s\n", name);
      perror (name);
      return -1;
    }
  return t_fd;
}

int 
main (int argc, char *argv[])
{
  if (argc < 2)
    {
      printf (usage_string, argv[0]);
      return -1;
    }
  signal (SIGINT, &signal_handler);
  oper_filename[0] = 0;
  strncpy (mixer_filename, DEFAULT_MIXER_FILE, STRINGLEN);
  strncpy (dsp_filename, DEFAULT_DSP_FILE, STRINGLEN);
  for (count = 0; count < SOUND_MIXER_NRDEVICES; count++)
    vol_requests[count].devnumber = -1;

  while (1)
    {
      value = getopt_long (argc, argv, "hivcles:b:d:a:r:m:p:f:t:", long_options, &optindex);
      if (value == -1)
	break;
      else
	switch (value)
	  {
	  case 's':
	    reqsamplerate = atoi (optarg);
	    actions |= AC_SETSAMPLERATE;
	    break;
	  case 'b':
	    reqbitrate = atoi (optarg);
	    actions |= AC_SETBITRATE;
	    break;
	  case 'd':
	    durationsecs = durationbytes = 0;
	    for (count = 0; optarg[count]; count++)
	      if (optarg[count] == 'b')
		{
		  optarg[count] = 0;
		  durationbytes = atoi (optarg);
		  durationsecs = -1;
		}
	    if (durationsecs == -1)
	      durationsecs = 0;
	    else
	      durationsecs = atoi (optarg);
	    actions |= AC_DURATION;
	    break;
	  case 'i':
	    actions |= AC_DISPINFO;
	    break;
	  case 'm':
	    strncpy (mixer_filename, optarg, STRINGLEN);
	    break;
	  case 'p':
	    strncpy (dsp_filename, optarg, STRINGLEN);
	    break;
	  case 'a':
	    for (count = 0; (count < SOUND_MIXER_NRDEVICES)
		 && strcmp (optarg, device_names[count]); count++);
	    if (count == SOUND_MIXER_NRDEVICES)
	      {
		printf ("error: device %s cannot be found\n", optarg);
		break;
	      }
	    reqaddsrcmask |= (1 << count);
	    actions |= AC_ADDRECSRC;
	    actions |= AC_SETRECSRC;
	    break;
	  case 'r':
	    for (count = 0; (count < SOUND_MIXER_NRDEVICES)
		 && strcmp (optarg, device_names[count]); count++);
	    if (count == SOUND_MIXER_NRDEVICES)
	      {
		printf ("error: device %s cannot be found\n", optarg);
		break;
	      }
	    reqremsrcmask |= (1 << count);
	    actions |= AC_REMRECSRC;
	    actions |= AC_SETRECSRC;
	    break;
	  case 'h':
	    actions |= AC_DISPHELP;
	    break;
	  case 'v':
	    actions |= AC_DISPVOLS;
	    break;
	  case 't':
	    // since we don't know what is the device file for the mixer,
	    // we need to build a request table for volume changes, 
	    // which will be used, later, after the getoptlong() loop

	    // find a valid slot
	    for (count2 = 0; (count2 < SOUND_MIXER_NRDEVICES) && (vol_requests[count2].devnumber != -1); count2++);
	    for (count = 0; (optarg[count] != '-') && (optarg[count]); count++);
	    memcpy (tempstring, optarg, count);
	    tempstring[count] = 0;
	    strcpy (tempstring2, tempstring);

	    tleft = tright = 0;
	    count++;		// skip the '-'

	    optarg += count;
	    for (count = 0; (optarg[count] != ':') && (optarg[count]); count++);
	    memcpy (tempstring, optarg, count);
	    tempstring[count] = 0;
	    if (!optarg[count])	// no ':', left = right

	      tleft = tright = atoi (tempstring);
	    else
	      {
		tleft = atoi (tempstring);
		count++;	// skip the    

		optarg += count;
		tright = atoi (optarg);
	      }
	    // ok, enough parsing

	    for (count = 0; (count < SOUND_MIXER_NRDEVICES)
		 && strcmp (tempstring2, device_names[count]); count++);
	    if (count == SOUND_MIXER_NRDEVICES)
	      {
		printf ("error: device %s cannot be found\n", tempstring2);
		break;
	      }
	    else
	      vol_requests[count2].devnumber = count;
	    vol_requests[count2].right = tright;
	    vol_requests[count2].left = tleft;
	    actions |= AC_SETVOLUMES;
	    break;
	  case 'f':
	    strncpy (oper_filename, optarg, STRINGLEN);
	    break;
	  case 'e':
	    actions |= AC_SETSTEREO;
	    break;
	  case 'c':
	    actions |= AC_RECORD;
	    break;
	  case 'l':
	    actions |= AC_PLAY;
	    break;
	  default:
	    break;
	  }
    }

  if ((audio_fd = safe_open (dsp_filename, O_RDWR)) == -1)
    return -1;
  if ((mixer_fd = safe_open (mixer_filename, O_RDWR)) == -1)
    return -1;

  if (ioctl (mixer_fd, SOUND_MIXER_READ_DEVMASK, &devmask) == -1)
    {
      printf ("error: cannot get device mask off mixer\n");
      return -1;
    }

  if (ioctl (mixer_fd, SOUND_MIXER_READ_RECSRC, &srcmask) == -1)
    {
      printf ("error: cannot get source mask off mixer\n");
      return -1;
    }

  if (ioctl (mixer_fd, SOUND_MIXER_READ_RECMASK, &recmask) == -1)
    {
      printf ("error: cannot get recording mask off mixer\n");
      return -1;
    }

  if (actions & AC_DISPHELP)
    {
      printf (usage_string, argv[0]);
      printf (help_string);
    }

  if (actions & AC_SETSTEREO)
    {
      stereo = 1;
      if (ioctl (audio_fd, SNDCTL_DSP_STEREO, &stereo) == -1)
	{
	  printf ("error: cannot set dsp into stereo mode");
	  perror ("ioctl");
	}
      else
	printf ("Stereo mode set on %s\n", dsp_filename);
    }

  if (actions & AC_SETSAMPLERATE)
    {
      printf ("Samplerate of %d hertz requested\n", reqsamplerate);
      if (ioctl (audio_fd, SNDCTL_DSP_SPEED, &reqsamplerate) == -1)
	printf ("error: can't set samplerate %d\n", reqsamplerate);
      else
	printf ("Samplerate set to %d hertz\n", reqsamplerate);
    }

  if (actions & AC_SETBITRATE)
    {
      if ((reqbitrate != 8) && (reqbitrate != 16))
	printf ("error: only bitrates of 8 and 16 accepted (%d invalid)\n", reqbitrate);
      else
	{
	  if ((ioctl (audio_fd, SNDCTL_DSP_SAMPLESIZE, &reqbitrate)) == -1)
	    printf ("error: cannot set a bitrate of %d\n", reqbitrate);
	  else
	    printf ("Bitrate set to %d\n", reqbitrate);
	}
    }

  if (actions & AC_SETVOLUMES)
    {
      for (count = 0; count < SOUND_MIXER_NRDEVICES; count++)
	if (vol_requests[count].devnumber != -1)
	  {
	    count2 = (vol_requests[count].right << 8)
	      | (vol_requests[count].left);
	    if (ioctl (audio_fd, MIXER_WRITE (vol_requests[count].devnumber), &count2) != -1)
	      printf ("Volume of %s set to left=%d, right=%d\n",
		      device_names[vol_requests[count].devnumber],
		      vol_requests[count].left,
		      vol_requests[count].right);
	  }
    }
  if (actions & AC_ADDRECSRC)
    {
      for (count = 0; count < SOUND_MIXER_NRDEVICES; count++)
	if ((1 << count) & reqaddsrcmask)
	  {
	    if (!((1 << count) & recmask))
	      printf ("error: device %s not supported for recording\n", device_names[count]);
	    else
	      srcmask |= (1 << count);
	  }
    }

  if (actions & AC_REMRECSRC)
    {
      for (count = 0; count < SOUND_MIXER_NRDEVICES; count++)
	if ((1 << count) & reqremsrcmask)
	  {
	    if (!((1 << count) & recmask))
	      printf ("error: device %s not supported for recording\n", device_names[count]);
	    else
	      srcmask &= ~(1 << count);
	  }
    }
  if (actions & AC_SETRECSRC)
    {
      if (ioctl (audio_fd, SOUND_MIXER_WRITE_RECSRC, &srcmask) == -1)
	{
	  printf ("error: couldn't write record sources\n");
	  perror ("ioctl");
	}
    }

  if (ioctl (audio_fd, SNDCTL_DSP_GETBLKSIZE, &audio_buffer_size) == -1)
    {
      printf ("error: cannot get block size off dsp\n");
      return -1;
    }

  bitrate = samplerate = 0;
  ioctl (audio_fd, SNDCTL_DSP_SAMPLESIZE, &bitrate);
  ioctl (audio_fd, SNDCTL_DSP_SPEED, &samplerate);

  if (actions & AC_DISPINFO)
    {
      printf ("Summary information for %s and %s:\n", mixer_filename, dsp_filename);
      // display supported devices
      printf ("Supported devices: ");
      for (count = 0; count < SOUND_MIXER_NRDEVICES; count++)
	{
	  if ((1 << count) & devmask)
	    printf ("%s ", device_names[count]);
	}
      printf ("\n");
      // display supported recording devices
      printf ("Supported recording devices: ");
      for (count = 0; count < SOUND_MIXER_NRDEVICES; count++)
	{
	  if ((1 << count) & recmask)
	    printf ("%s ", device_names[count]);
	}
      printf ("\n");

      // display current recording devices REFRESH

      printf ("Current recording devices: ");
      for (count = 0; count < SOUND_MIXER_NRDEVICES; count++)
	{
	  if ((1 << count) & srcmask)
	    printf ("%s ", device_names[count]);
	}
      printf ("\n");

      // display bitrate, sampleheight of /dev/dsp
      printf ("Actual bitrate: %d bits per sample\n", bitrate);
      printf ("Actual samplerate: %d hertz\n", samplerate);
      printf ("Audio buffer size: %d bytes\n", audio_buffer_size);
      printf ("\n");
    }

  if (actions & AC_DISPVOLS)
    {
      int temp;
      printf ("Volume information for %s:\n", mixer_filename);
      for (count = 0; count < SOUND_MIXER_NRDEVICES; count++)
	{
	  if (ioctl (mixer_fd, MIXER_READ (count), &temp) != -1)
	    printf ("\t%s, left=%d right=%d\n", device_names[count],
		    temp & 0xff, (temp >> 8) & 0xff);
	}
    }

  if ((actions & AC_RECORD) && (actions & AC_PLAY))
    {
      actions &= ~(AC_RECORD | AC_PLAY);
      printf ("error: cannot record and play at the same time\n");
    }

  if (actions & (AC_RECORD | AC_PLAY))
    if ((audio_buffer = malloc (audio_buffer_size)) == NULL)
      {
	printf ("error: cannot alocate audio buffer (%d bytes)\n", audio_buffer_size);
	exit (-1);		// bawh

      }
  if (actions & AC_DURATION)
    {
      if (!durationbytes)
	durationbytes = durationsecs * samplerate * (bitrate >> 3) * (stereo ? 2 : 1);
      if (!durationsecs)
	durationsecs = durationbytes / (samplerate * (bitrate >> 3) * (stereo ? 2 : 1));
    }

  if (actions & AC_RECORD)
    {
      int numread;
      int total;
      if (!oper_filename[0])
	oper_fd = 1;		// stdout

      else if ((oper_fd = safe_open (oper_filename, O_WRONLY | O_CREAT)) == -1)
	goto endrec;
      numread = -1;
      if (actions & AC_DURATION)
	printf ("Recording for %f seconds (%d bytes)\n", durationsecs, durationbytes);
      while (oper_running && numread)
	{
	  numread = read (audio_fd, audio_buffer, audio_buffer_size);
	  if (actions & AC_DURATION)
	    {
	      if ((durationbytes - numread) > 0)
		{
		  durationbytes -= numread;
		  write (oper_fd, audio_buffer, numread);
		  total+=numread;
		}
	      else if ((durationbytes - numread <= 0))
		{
		  write (oper_fd, audio_buffer, durationbytes);
		  total+=durationbytes;
		  break;
		}
	    }
	  else
	    {
	    write (oper_fd, audio_buffer, audio_buffer_size);
	    total+=audio_buffer_size;
	    }

	}
      close (oper_fd);
    endrec:
    }
  else if (actions & AC_PLAY)
    {
      int numread;
      if (!oper_filename[0])
	oper_fd = 0;		// stdin

      else if ((oper_fd = safe_open (oper_filename, O_RDONLY)) == -1)
	goto endplay;
      numread = -1;
      total=0;
      if (actions & AC_DURATION)
	printf ("Playing for %f seconds (%d bytes)\n", durationsecs, durationbytes);
      while (oper_running && numread)
	{
	  numread = read (oper_fd, audio_buffer, audio_buffer_size);
	  if (actions & AC_DURATION)
	    {
	      if ((durationbytes - numread) > 0)
		{
		  durationbytes -= numread;
		  write (audio_fd, audio_buffer, numread);
		  total+=numread;
		}
	      else if ((durationbytes - numread) <= 0)
		{
		  write (audio_fd, audio_buffer, durationbytes);
		  total+=durationbytes;
		  break;
		}
	    }
	  else
	    {
	    write (audio_fd, audio_buffer, audio_buffer_size);
	    total+=audio_buffer_size;
	    }
	}
      close (oper_fd);
    endplay:
    }
  if (actions & AC_RECORD)
   printf("Recorded %d bytes, for %d seconds\n", total, (int)total/(samplerate * (bitrate >> 3) * (stereo ? 2 : 1)));
  if (actions & AC_PLAY)
   printf("Played %d bytes, for %d seconds\n", total, (int)total/(samplerate * (bitrate >> 3) * (stereo ? 2 : 1)));
  if (!oper_running)
    printf ("Caught a SIGINT, terminating\n");
  close (mixer_fd);
  close (audio_fd);
  return 0;
}
